/*

 ideservd.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>
 Win32 port by
  Josef Soucek <josef.soucek@ide64.org>
 Serial support by
  Borsanyi Attila <bigfoot@axelero.hu>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ideservd.h"
#include "config.h"
#include <errno.h>
#ifndef WIN32		//*NIX
#include <pwd.h>
#include <grp.h>
#include <sys/wait.h>
#else			//WIN32
#include <windows.h>
#include "resource.h"
#endif 			//ALL
#include <signal.h>
#include <unistd.h>
#ifdef __DJGPP__
#define chroot(a) 0
#endif
#include <string.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <locale.h>
#include "path.h"
#include "partition.h"
#include "driver.h"
#include "arguments.h"
#ifndef __DJGPP__
#include "usb.h"
#include "eth.h"
#include "vice.h"
#endif
#include "rs232.h"
#ifndef OSX
#include "x1541.h"
#include "pc64.h"
#endif

#include "crc8.h"
#include "compat.h"
#include "normal.h"
#include "log.h"
#ifdef __MINGW32__
#define mkdir(a,b) mkdir (a)
#endif

static const struct io_ops_s *io_ops;
static Log *lf;

static Arguments arguments;
static struct buffer_s buff[16];

#ifdef WIN32
/* Prototypes */
static INT_PTR CALLBACK MyDlgProc(HWND,UINT,WPARAM,LPARAM);
static long WINAPI PCLinkLoop(long);
/* Global variable */
static NOTIFYICONDATA g_notifyIconData;
#endif

static void terminate(int UNUSED(x)) {
    if (arguments.mode != M_ETHERNET) io_ops->shutdown();
    if (lf) {
	log_print(lf, "Terminated\n");
	log_close(lf);
    }
    exit(0);
}

void seterror(int c,int i1) {
    const char *msg="UNKNOWN ERROR";

    switch (c) {
    case 00:msg="OK";break;
    case 01:msg="FILES SCRATCHED";break;
    case 02:msg="PARTITION SELECTED";break;
    case 23:if (i1==0) msg="READ ERROR"; else msg="CRC ERROR";i1=0;break;
    case 26:msg="WRITE PROTECT ON";break;
    case 27:msg="ACCESS DENIED";break;
    case 31:msg="UNKNOWN COMMAND";break;
    case 30:
    case 33:
    case 34:msg="SYNTAX ERROR";break;
    case 39:msg="PATH NOT FOUND";break;
    case 41:msg="FRAME ERROR";break;
    case 42:msg="CRC ERROR";break;
    case 62:msg="FILE NOT FOUND";break;
    case 63:msg="FILE EXISTS";break;
    case 64:msg="FILE TYPE MISMATCH";break;
    case 70:msg="NO CHANNEL";break;
    case 72:msg="PARTITION FULL";break;
    case 73:msg="IDESERV " VERSION2 " PCLINK";break;
    case 77:msg="SELECTED PARTITION ILLEGAL";break;
    }
    buff[15].length = sprintf((char *)buff[15].data,"%02d, %s,%03d,000,000,000",c,msg,i1);
    buff[15].pointer = 0;
}

volatile int tout;
static void timeout(int UNUSED(x)) {
    log_print(lf, "Timeout\n");
    io_ops->clean();
    tout=1;
}

int errtochannel15(int i) {
    if (i)
        switch (errno) {
        case EISDIR:
        case ENOTDIR:
        case ENOENT:seterror(62,0);return 64;
        case EEXIST:seterror(63,0);return 63;
        case EPERM:
        case EACCES:seterror(27,0);return 27;
        case ENOTEMPTY:seterror(1,0);return 1;
        case ENOSPC:seterror(72,0);return 72;
        case EROFS:seterror(26,0);return 26;
        case EIO:seterror(23,0);return 23;
        default:seterror(80,errno); return 0;
        }
    seterror(0,0);
    return 0;
}

static void do_chdir(const c64c_t *cmd) {
    c64c_t name[17];
    const c64c_t *outname;
    char outpath[1020];
    int found=0;
    partition_t outpart = 0;
    directory_entry_t dirent;

    if (!strchr((char *)cmd, ':')) {
        if (!strcmp((char *)cmd, "_")) cmd=(c64c_t *)":_";
        else if (cmd[0] != '/') { seterror(30,0); goto vege2; }
        else cmd++;
    }

    if (!(found=resolv_path(cmd, outpath, &outname, &outpart, lf, arguments.nameconversion))) {seterror(39,0);goto vege2;}

    if (!outname[0]) goto vege2;
    if (!strcmp((const char *)outname,"_") || !strcmp((const char *)outname, "..")) {
        char *o = strrchr(outpath, '/');
        if (o) *o=0; else outpath[0]=0;
        goto vege2;
    }
    if (!strcmp((const char *)outname, ".")) goto vege2;
    found = 0;
    convertfilename(outname, ',', name, NULL, NULL);

    if (directory_open(outpath, arguments.nameconversion, 0)) {
        log_print(lf, "Couldn't open the directory \"%s\": %s(%d)\n", outpath, strerror(errno), errno);
        errtochannel15(1);
        return;
    }

    while (directory_read(&dirent)) 
    {
        if ((dirent.attrib & A_ANY)!=A_DIR) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name, name)) continue;

        strcpy(outpath, directory_path());
        found=1;
        break;
    }
    directory_close();
vege2:
    if (found) {
        partition_set_path(outpart, outpath);
        seterror(0,0);
    } else seterror(39,0);
    if (arguments.verbose) log_print(lf, "Change directory \"%s\"\n", outpath[0] ? outpath : "/");
}

static void do_mkdir(const c64c_t *cmd) {
    char outpath[1020],lname[1000];
    int found=0;
    c64c_t name[17];
    const c64c_t *outname;
    directory_entry_t dirent;

    if (!strchr((char *)cmd, ':')) {seterror(30,0);found=0;outpath[0]=0;goto vege2;}

    found=0;
    if (!resolv_path(cmd, outpath, &outname, NULL, lf, arguments.nameconversion)) {seterror(39,0);goto vege2;}

    convertfilename(outname, ',', name, NULL, NULL);

    if (directory_open(outpath, arguments.nameconversion, 0)) {
        log_print(lf, "Couldn't open the directory \"%s\": %s(%d)\n", outpath, strerror(errno), errno);
        errtochannel15(1);
        return;
    }

    while (directory_read(&dirent))
    {
        if ((dirent.attrib & A_ANY)!=A_DIR) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name, name)) continue;

        strncpy(lname, directory_filename(), 999);

        found=1;
        break;
    }
    directory_close();

    if (!found) {
        mbstate_t ps;
        int f, f2;
        memset(&ps, 0, sizeof(ps));
        for (f = f2 = 0; name[f] && f < 16; f++) {
            f2 += c64toascii(lname + f2, name[f], &ps);
        }
        lname[f2]=0;
    }

    if (outpath[0]) strcat(outpath,"/");
    strcat(outpath,lname);

    if (found)
        seterror(63,0);
    else {
        int f=0;

        while (lname[f]) {
            if (lname[f]=='*' || lname[f]=='?' || lname[f]==':' || lname[f]=='=') {
                seterror(33,0);goto vege2;
            }
            f++;
        }
        errtochannel15(mkdir(outpath,0777));
    }
vege2:
    if (!name[0]) {
        seterror(34,0);
    }
    if (arguments.verbose) log_print(lf, "Make directory \"%s\"\n", outpath[0] ? outpath : "/");
}

static void do_scratch(const c64c_t *cmd, int smode) {
    char outpath[1000];
    int found=0;
    c64c_t name[17], type[4]={'*',0};
    const c64c_t *outname;
    directory_entry_t dirent;

    if (!strchr((char *)cmd, ':')) {seterror(30,0);found=0;outpath[0]=0;goto vege;}

    found=0;
    if (!resolv_path(cmd, outpath, &outname, NULL, lf, arguments.nameconversion)) {seterror(39,0);goto vege;}

    convertfilename(outname, '=', name, type, NULL);

    if (directory_open(outpath, arguments.nameconversion, 0)) {
        log_print(lf, "Couldn't open the directory \"%s\": %s(%d)\n", outpath, strerror(errno), errno);
        errtochannel15(1);
        return;
    }

    while (directory_read(&dirent))
    {
        const char *path;
        if (smode) {
            if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_DIR | A_CLOSED)) continue;
        } else {
            if ((dirent.attrib & A_ANY)==A_DIR) continue;
        }

        if (!(dirent.attrib & A_DELETEABLE)) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name, name)) continue;
        if (!smode) {
            dirent.time[0]=0;
            if (!matchname(dirent.filetype,type)) continue;
        }

        path = directory_path();
        found=1;
        if (smode) {
            if (arguments.verbose) log_print(lf, "Remove directory \"%s\"\n", path);
            errtochannel15(rmdir(path));
        } else {
            errtochannel15(unlink(path));
            if (arguments.verbose) log_print(lf, "Remove \"%s\"\n", path);
        }
    }
    directory_close();
    if (!found) seterror(1,0);
vege:
    if (!found) {
        if (smode) {
            if (arguments.verbose) log_print(lf, "Remove directory \"%s\"\n", outpath);
        } else {
            if (arguments.verbose) log_print(lf, "Remove \"%s\"\n", outpath);
        }
    }

    if (!name[0]) {
        seterror(34,0);
    }
}

void commandchannel(const c64c_t *s) {
    if (s[0]=='C' && s[1]=='D') {
        do_chdir(s+2);
    } else if (s[0]=='M' && s[1]=='D') {
        do_mkdir(s);
    } else if (s[0]=='R' && s[1]=='D') {
        do_scratch(s,1);
    } else if (s[0]=='S' && s[1]!='-') {
        do_scratch(s,0);
    } else if (s[0]=='C' && s[1]==0xd0) {
        if (arguments.verbose) log_print(lf, "Change to partition %d\n", s[2]);
        seterror(partition_select(s[2]),s[2]);
    } else if (s[0]=='C' && s[1]=='P') {
        partition_t part = 0;
        int i = 2;
        if (!s[2]) part = partition_get_current(); else {
            while ((s[i] ^ 0x30) < 10) part = part * 10 + (s[i++] ^ 0x30);
        }
        if (arguments.verbose) log_print(lf,"Change to partition %d\n", part);
        seterror(partition_select(part), part);
    } else if (s[0]=='U' && ((s[1] & 0x0f)==9 || (s[1] & 0x0f)==10)) {
        seterror(73,0);
        if (arguments.verbose) log_print(lf,"Identify device\n");
    } else if (s[0]=='I') {
        seterror(0,0);
        if (arguments.verbose) log_print(lf,"Initialize\n");
    } else if (s[0]=='V') {
        seterror(0,0);
        if (arguments.verbose) log_print(lf,"Validate\n");
    } else if (s[0]=='M' && s[1]=='-' && s[2]=='R') { 
        int i,j,k;
        j=s[5];if (j>50) j=50;
        k=s[3];
	for (i=0;i<j;i++) buff[15].data[i]="IDE64 CARTRIDGE "[k++ & 15];
	buff[15].data[i]=0;
        if (arguments.verbose) log_print(lf, "Memory read $%04x %02x\n", s[3] | (s[4] << 8), s[5]);
    } else if (s[0]) {
        seterror(31,0);
    }
}

static int init(int argc, char *argv[])
{
#ifndef WIN32
    struct passwd *user2 = NULL;
    struct group *group2 = NULL;
#endif
    const char *init_dev = NULL;
    int init_int = 0;
    int lastfail = 0;

    setlocale(LC_ALL, "");

    testarg(&arguments, argc, argv);
    partition_create(1, (c64c_t *)"PARTITION 1");
    partition_select(1);

    lf = log_open(arguments.log);
    log_print(lf, "ideservd " VERSION "\n");

    arguments.mode = M_NONE;
    if (arguments.mode_name) {
        const struct {
            const char *name;
            enum e_modes mode;
        } *i, modes[] = {
#ifdef _X1541_H
            {"x1541", M_X1541}, {"xe1541", M_XE1541}, {"xm1541", M_XM1541}, {"xa1541", M_XA1541},
#endif
#ifdef _PC64_H
            {"pc64", M_PC64}, {"pc64s", M_PC64S},
#endif
#ifdef _RS232_H
            {"rs232", M_RS232}, {"rs232s", M_RS232S},
#endif
#ifdef _USB_H
            {"usb", M_USB},
#endif
#ifdef _VICE_H
            {"vice", M_VICE},
#endif
#ifdef _ETH_H
            {"eth", M_ETHERNET},
#endif
            {NULL, M_NONE}
        };
        i = modes;
        while (i->name) {
            if (!strcasecmp(arguments.mode_name, i->name)) {
                arguments.mode = i->mode;
                break;
            }
            i++;
        }
    } else {
#ifdef _USB_H
        arguments.mode = M_USB;
#else
        arguments.mode = M_X1541;
#endif
    }

    switch (arguments.mode) {
#ifdef _X1541_H
    case M_X1541:
    case M_XE1541:
    case M_XM1541:
    case M_XA1541:
        {
            const char *cable = "X";
            io_ops=&x1541_io_ops;
            if (arguments.lptport == 0) arguments.lptport = DEFAULT_IOPORT;
            init_int = arguments.lptport & 0xffff;
            if (arguments.hog) init_int |= X1541_HOG;
            if (arguments.device) init_int |= X1541_DEVICE;
            if (arguments.mode == M_XE1541) {init_int |= X1541_XE1541; cable = "XE";}
            if (arguments.mode == M_XM1541) {init_int |= X1541_XM1541; cable = "XM";}
            if (arguments.mode == M_XA1541) {init_int |= X1541_XA1541; cable = "XA";}
            if (arguments.device) {
                log_print(lf, "Using %s driver with %s1541 cable on device %s\n", io_ops->drivername(), cable, arguments.device);
            } else {
                log_print(lf, "Using %s driver with %s1541 cable at port %x\n", io_ops->drivername(), cable, arguments.lptport);
            }
            init_dev = arguments.device;
        }
        break;
#endif
#ifdef _PC64_H
    case M_PC64:
    case M_PC64S:
        {
            io_ops=&pc64_io_ops;
            if (arguments.lptport == 0) arguments.lptport = DEFAULT_IOPORT;
            init_int = arguments.lptport & 0xffff;
            if (arguments.hog) init_int |= PC64_HOG;
            if (arguments.device) init_int |= PC64_DEVICE;
            if (arguments.mode == M_PC64S) init_int |= PC64_SLOW;
            if (arguments.device) {
                log_print(lf, "Using %s driver on device %s\n", io_ops->drivername(), arguments.device);
            } else { 
                log_print(lf, "Using %s driver at port %x\n", io_ops->drivername(), arguments.lptport);
            }
            init_dev = arguments.device;
        }
        break;
#endif
#ifdef _RS232_H
    case M_RS232:
    case M_RS232S:
        io_ops=&rs232_io_ops;
        if (!arguments.device) arguments.device = DEFAULT_COMPORT;
        log_print(lf, "Using %s driver on device %s\n", io_ops->drivername(), arguments.device);
        init_dev = arguments.device;
        init_int = (arguments.mode == M_RS232S);
        break;
#endif
#ifdef _USB_H
    case M_USB:
        io_ops=&usb_io_ops;
        init_dev = arguments.device;
        log_print(lf, "Using %s driver for device %s\n", io_ops->drivername(), arguments.device ? arguments.device : "*");
        break;
#endif
#ifdef _VICE_H
    case M_VICE:
        io_ops=&vice_io_ops;
        init_dev = arguments.sin_addr;
        init_int = arguments.network;
        if (arguments.network) {
            log_print(lf, "Using %s driver connecting to %s port %d\n", io_ops->drivername(), arguments.sin_addr ? arguments.sin_addr : "127.0.0.1", arguments.network);
        } else {
            log_print(lf, "Using %s driver connecting to %s\n", io_ops->drivername(), arguments.sin_addr ? arguments.sin_addr : "127.0.0.1");
        }
        break;
#endif
#ifdef _ETH_H
    case M_ETHERNET:
        io_ops = &eth_io_ops;
        init_dev = arguments.sin_addr;
        init_int = arguments.network;
        if (arguments.sin_addr) {
            log_print(lf, "Using %s driver connecting to %s network %d\n", io_ops->drivername(), arguments.sin_addr, arguments.network);
        } else {
            log_print(lf, "Using %s driver connecting to 127.0.0.1\n", io_ops->drivername());
        }
        break;
#endif
    default: 
        log_print(lf, "No driver for unknown mode \"%s\"\n", arguments.mode_name);
        log_close(lf);
        exit(EXIT_FAILURE);
    }

#ifndef WIN32
    {
        pid_t pid;

        if (arguments.background) if (fork()) {
            log_print(lf, "Forking into background\n");
            log_close(lf);
            exit(EXIT_SUCCESS);
        }
        for (;;) {
            pid=fork();
            if (pid) {
                int status;
                usleep(1000000);
                waitpid(pid, &status, 0);
                lastfail = (signed char)WEXITSTATUS(status);
            } else break;
        }
    }
#endif

    do {
        int thisfail = io_ops->initialize(init_dev, init_int, lf, lastfail);
        int change = thisfail != lastfail;
        if (!thisfail) break;
        lastfail = thisfail;
        if (change) {
#ifndef WIN32
            exit(thisfail);
#else
            break;
#endif
        }
        usleep(1000000);
    } while (1);

#ifdef WIN32
    if (chdir(arguments.root ? arguments.root : "/cygdrive")) {
        if (lastfail != 2) log_print(lf, "Changing dir to \"%s\" failed: %s(%d)\n", arguments.root ? arguments.root : "/cygdrive", strerror(errno), errno);
        if (arguments.root) {
            log_close(lf);
            exit(2);
        }
    }
#else
    errno = 0;
    if (arguments.priority && nice(arguments.priority) == -1 && errno != 0) {
        log_print(lf, "Couldn't adjust priority: %s(%d)\n", strerror(errno), errno);
    }
    if (!arguments.user || !arguments.group) {
        uid_t uid = getuid(), euid = geteuid();
        if (uid <= 0 || uid != euid) {
            if (!arguments.user) arguments.user = "nobody";
            if (!arguments.group) arguments.group = "nogroup";
        }
    }
    if (arguments.user && !(user2=getpwnam(arguments.user))) {
        if (lastfail != 3) log_print(lf, "User \"%s\" does not exist\n", arguments.user);
        log_close(lf);
        exit(3);
    }
    if (arguments.group && !(group2=getgrnam(arguments.group))) {
        if (lastfail != 4) log_print(lf, "Group \"%s\" does not exist\n", arguments.group);
        log_close(lf);
        exit(4);
    }
// chroot crashes on WIN32 with current cygwin1.dll without registry entry:
// [HKEY_LOCAL_MACHINE\SOFTWARE\Cygnus Solutions\Cygwin\mounts v2\/]
// "native"="C:\\"
    if (arguments.root) {
        uid_t uid = getuid(), euid = geteuid();
        if (chdir(arguments.root)) {
            if (lastfail != 5) log_print(lf, "Changing dir to \"%s\" failed: %s(%d)\n", arguments.root, strerror(errno), errno);
            log_close(lf);
            exit(5);
        } else {
            if (chroot(arguments.root) || chdir("/")) {
                if (uid <= 0 || uid != euid) {
                    if (lastfail != 6) log_print(lf, "Chrooting failed: %s(%d)\n", strerror(errno), errno);
                    log_close(lf);
                    exit(6);
                }
                log_print(lf, "Chrooting failed: %s(%d)\n", strerror(errno), errno);
            }
        }
    }
    if ((group2 && setgid(group2->gr_gid)) || (user2 && setuid(user2->pw_uid))) {
        if (lastfail != 7) log_print(lf, "Could not drop privileges: %s(%d)\n", strerror(errno), errno);
        log_close(lf);
        exit(7);
    }
#endif
    return 0;
}

#ifdef WIN32
INT_PTR CALLBACK MyDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
/*  int i;
  HWND hwnd_select;

  const char *pclinkmodes[] = {
            "x1541",
            "xe1541",
            "xm1541",
            "xa1541",
            "pc64",
            "pc64s",
            "rs232",
            "rs232s",
            "usb",
            "eth" };*/

  switch(message) 
  {
//   case WM_INITDIALOG:
//     hwnd_select = GetDlgItem(hwnd, IDC_SELECT);
//     for(i=0; i<11; i++) SendMessage(hwnd_select, CB_ADDSTRING, 0, (LPCTSTR)pclinkmodes[i]);
//     SendMessage(hwnd_select, CB_SETCURSEL, 8, 0); // from 0
//     return TRUE;
//   break;

   case WM_COMMAND:
	if (LOWORD(wParam) == IDC_EXIT || LOWORD(wParam) == IDCANCEL)
	{
		PostQuitMessage(0);
		EndDialog(hwnd,TRUE);
		return TRUE;
        }

	if (LOWORD(wParam) == IDC_BUTTON)
	{
	  ShowWindow(hwnd, SW_HIDE);
	}
   break;

//   case WM_SYSCOMMAND:
//	if (LOWORD(wParam) == SC_MINIMIZE)
//	{
//	   ShowWindow(hwnd, SW_HIDE);
//	}
//   break;

   case WM_USER:
	if (lParam == WM_RBUTTONDOWN || lParam == WM_LBUTTONDOWN)
        {
	   ShowWindow(hwnd, SW_SHOW);
   	}
   break;
  }
return FALSE;
}
#endif

int main(int argc,char *argv[])
// int WINAPI WinMain (HINSTANCE hInstance,
//                    HINSTANCE hPrevInstance,
//                    LPSTR szCmdLine,
//                    int iCmdShow)
{
    int b;
#ifdef WIN32
    int lastfail = init(argc, argv);
#else
    init(argc, argv);
#endif
    signal(SIGALRM, timeout);
    signal(SIGTERM, terminate);
    signal(SIGINT, terminate);
    for (b=0;b<15;b++) buff[b].mode = CM_CLOSED;
    buff[15].mode = CM_ERR;
    if (!buff[15].data) {
        buff[15].data = (unsigned char *)malloc(256);
        if (!buff[15].data) exit(1); // out of memory?
    }
    seterror(73,0);

#ifdef WIN32
    HWND hwnd;
    MSG msg;

    hwnd=CreateDialog(NULL, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC)MyDlgProc);
    ShowWindow(hwnd, SW_SHOW);
 
// TRAY ICON

    memset(&g_notifyIconData, 0, sizeof(NOTIFYICONDATA));
    g_notifyIconData.cbSize = sizeof(NOTIFYICONDATA);
    g_notifyIconData.hWnd = hwnd;
    g_notifyIconData.uID = 0;
    g_notifyIconData.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
    g_notifyIconData.uCallbackMessage = WM_USER; 
    g_notifyIconData.hIcon = (HICON)LoadImage( NULL, TEXT("pclink.ico"), IMAGE_ICON, 0, 0, LR_LOADFROMFILE);
    strcpy(g_notifyIconData.szTip, "IDE64 PCLink");
    Shell_NotifyIcon(NIM_ADD, &g_notifyIconData);
    
    HANDLE hThread;
    DWORD iID;
    hThread = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)PCLinkLoop,(LPVOID)lastfail,0,&iID);
    
    while(GetMessage (&msg,NULL,0,0)!=0) {
//       TranslateMessage(&msg);  // Only if there is text input
       DispatchMessage(&msg);
    }
    TerminateThread(hThread,0);
    Shell_NotifyIcon(NIM_DELETE, &g_notifyIconData);
    terminate(0);
    return 0;
}

long WINAPI PCLinkLoop(long lParam)
{
    int b;
    int lastfail = (int)lParam;
#endif //WIN32

    for (;;) {
        if (arguments.verbose) log_print(lf,"ready\n");
        log_flush(lf);
    start2:
        crc_clear();
        b=io_ops->wait();
        if (b<0) {
#ifdef WIN32
            usleep(1000000);
            lastfail = io_ops->reinitialize(lf, lastfail);
            b=0;
#else
            terminate(0);
#endif
        }

        switch (b) {
        case 0x00:goto start2;
        case 'N':
        case 0xCE:b=openfile(io_ops, buff, lf, &arguments, b == 0xCE);break;
        case 'G':
        case 0xC7:b=readfile(io_ops, buff, lf, &arguments, b == 0xC7);break;
        case 'S':
        case 0xD3:b=writefile(io_ops, buff, lf, &arguments, b == 0xD3);break;
        case 'D':
        case 0xC4:b=closefile(io_ops, buff, lf, &arguments, b == 0xC4);break;
        case 'I':
        case 0xC9:b=statuserror(io_ops, (char *)buff[15].data, lf, &arguments, b == 0xC9);break;
        case 'O':
        case 0xCF:b=openfile_compat(io_ops, buff, lf, &arguments, b == 0xCF);break;
        case 'R':
        case 0xD2:b=readfile_compat(io_ops, buff, lf, &arguments, b == 0xD2);break;
        case 'W':
        case 0xD7:b=writefile_compat(io_ops, buff, lf, &arguments, b == 0xD7);break;
        case 'C':
        case 0xC3:b=closefile_compat(io_ops, buff, lf, &arguments, b == 0xC3);break;
        default:log_print(lf, "Unknown command %02X\n",b);
        }
        if (b) io_ops->clean();
    }
    return 0;
}
